use std::collections::HashMap;

use super::Solution;

pub struct Day14 {}

impl Solution for Day14 {
    fn part1(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        let mut rock_formation = input
            .iter()
            .map(|r| r.chars().collect::<Vec<_>>())
            .collect::<Vec<_>>();

        self.pull_north(&mut rock_formation);

        Ok(Box::new(self.get_load(&rock_formation)))
    }

    fn part2(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        let mut rock_formation = input
            .iter()
            .map(|r| r.chars().collect::<Vec<_>>())
            .collect::<Vec<_>>();

        let mut rotations: HashMap<Vec<Vec<char>>, usize> = HashMap::new();
        for i in 1..=1_000_000_000 {
            for _ in 0..4 {
                self.pull_north(&mut rock_formation);
                rock_formation = self.rotate(&rock_formation);
            }

            if let Some(rotation) = rotations.insert(rock_formation.clone(), i) {
                if (1_000_000_000 - i) % (i - rotation) == 0 {
                    break;
                }
            }
        }

        Ok(Box::new(self.get_load(&rock_formation)))
    }

    fn get_day(&self) -> u8 {
        14
    }
}

impl Day14 {
    // Function that pulls all of the 'O's as far north as they will go
    fn pull_north(&self, cave: &mut Vec<Vec<char>>) {
        let mut complete = false;

        while !complete {
            complete = true;
            for row in 1..cave.len() {
                for col in 0..cave[0].len() {
                    if cave[row][col] == 'O' && cave[row - 1][col] == '.' {
                        cave[row][col] = '.';
                        cave[row - 1][col] = 'O';
                        complete = false;
                    }
                }
            }
        }
    }

    // Function that calculates the load on the north beams of the cave
    fn get_load(&self, cave: &Vec<Vec<char>>) -> u32 {
        (0..cave.len())
            .map(|row| {
                (0..cave[0].len())
                    .filter(|col| cave[row][*col] == 'O')
                    .map(|_| cave.len() - row)
                    .sum::<usize>()
            })
            .sum::<usize>() as u32
    }

    // Function that rotates the given vector through 90 degrees clockwise
    fn rotate(&self, cave: &Vec<Vec<char>>) -> Vec<Vec<char>> {
        let mut new_cave = vec![vec![' '; cave.len()]; cave[0].len()];
        for row in 0..cave.len() {
            for col in 0..cave[0].len() {
                new_cave[col][cave.len() - row - 1] = cave[row][col];
            }
        }
        new_cave
    }
}

/// Test from puzzle input
#[cfg(test)]
mod test {
    use super::*;
    use crate::*;

    #[test]
    fn part1() {
        let challenge = day14::Day14 {};

        //Complete the Challenge
        let answer = challenge
            .part1(
                utils::get_input(challenge.get_day(), utils::InputType::Test1)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "136");
    }

    #[test]
    fn part2() {
        let challenge = day14::Day14 {};

        //Complete the Challenge
        let answer = challenge
            .part2(
                utils::get_input(challenge.get_day(), utils::InputType::Test2)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "64");
    }

    #[test]
    fn rotation_test() {
        let challenge = day14::Day14 {};

        let mut arr = vec![
            vec!['a', 'b', 'c', 'd'],
            vec!['d', 'e', 'f', 'g'],
            vec!['h', 'i', 'j', 'k'],
            vec!['l', 'm', 'n', 'o'],
        ];

        let ans = arr.clone();
        for _ in 0..4 {
            arr = challenge.rotate(&arr);
        }

        assert_eq!(arr, ans);
    }
}
