use std::str;

use super::Solution;

pub struct Day15 {}

impl Solution for Day15 {
    fn part1(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        let pre_hashes = input[0].split(',').collect::<Vec<_>>();

        let ans = pre_hashes
            .iter()
            .map(|h| self.hash(h, 17, 256))
            .sum::<u32>();
        Ok(Box::new(ans))
    }
    fn part2(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        let pre_hashes = input[0].split(',').collect::<Vec<_>>();
        let mut boxes: Vec<Vec<(String, u32)>> = vec![vec![]; 256];

        pre_hashes
            .iter()
            .for_each(|ins| self.sort_lenses(&mut boxes, ins));

        let ans = boxes
            .iter()
            .enumerate()
            .map(|(box_num, lenses)| {
                lenses
                    .iter()
                    .enumerate()
                    .map(|(lens_num, (_, focal_length))| {
                        self.get_focussing_power(box_num + 1, lens_num + 1, *focal_length)
                    })
                    .sum::<u32>()
            })
            .sum::<u32>();

        Ok(Box::new(ans))
    }

    fn get_day(&self) -> u8 {
        15
    }
}

impl Day15 {
    /// function to generate the predefined hash outlined in the challenge
    fn hash(&self, input: &str, seed: u32, divisor: u32) -> u32 {
        let mut hash = 0u32;

        for c in input.bytes() {
            hash += c as u32;
            hash *= seed;
            hash = hash % divisor;
        }

        hash
    }

    /// Sort the lenses in the boxes based in instructions given
    fn sort_lenses(&self, boxes: &mut Vec<Vec<(String, u32)>>, instruction: &str) {
        let bytes = instruction.as_bytes();

        // '-' Logic or '=' Logic
        if bytes[bytes.len() - 1] == b'-' {
            let label = str::from_utf8(&bytes[..bytes.len() - 1 as usize]).unwrap();
            let hash = self.hash(label, 17, 256);
            boxes[hash as usize].retain(|(lens, _)| lens != label);
        } else {
            let mut parts = instruction.split("=");
            let label = parts.next().unwrap();
            let focal_length = parts.next().unwrap().parse::<u32>().unwrap();

            let hash = self.hash(label, 17, 256);

            let old_position = boxes[hash as usize]
                .iter()
                .position(|(existing_label, _)| existing_label == label);
            if let Some(old_position) = old_position {
                boxes[hash as usize][old_position] = (label.to_string(), focal_length);
            } else {
                boxes[hash as usize].push((label.to_string(), focal_length));
            }
        }
    }

    /// Function that determines the focussing power of a lens given its box,
    /// position and focal length
    fn get_focussing_power(&self, box_num: usize, slot_num: usize, focal_length: u32) -> u32 {
        box_num as u32 * slot_num as u32 * focal_length
    }
}

/// Test from puzzle input
#[cfg(test)]
mod test {
    use super::*;
    use crate::*;

    #[test]
    fn part1() {
        let challenge = day15::Day15 {};

        //Complete the Challenge
        let answer = challenge
            .part1(
                utils::get_input(challenge.get_day(), utils::InputType::Test1)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "1320");
    }

    #[test]
    fn part2() {
        let challenge = day15::Day15 {};

        //Complete the Challenge
        let answer = challenge
            .part2(
                utils::get_input(challenge.get_day(), utils::InputType::Test2)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "145");
    }

    #[test]
    fn hash_test() {
        let challenge = day15::Day15 {};

        // Test the hash
        assert_eq!(challenge.hash("HASH", 17, 256), 52);
    }
}
