use super::Solution;

pub struct Day16 {}

#[derive(Clone, Copy, Debug)]
enum Direction {
    Up,
    Right,
    Down,
    Left
}

impl Solution for Day16 {
    fn part1(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        let grid = input.iter().map(|l| l.as_bytes()).collect::<Vec<_>>();

        Ok(Box::new(self.traverse_grid(&grid, (0, 0, Direction::Right))))
    }

    fn part2(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        let grid = input.iter().map(|l| l.as_bytes()).collect::<Vec<_>>();

        // Itterate through every edge piece and record the max energy from a list of all of the combinations
        // O(n^2) :( Can't think of a better way of processing this
        // r: row, c: col
        let ans = (0..grid.len()).flat_map(|r| [(r,0,Direction::Right), (r,grid[0].len()-1,Direction::Left)])
            .chain((0..grid[0].len()).flat_map(|c| [(0,c,Direction::Down), (grid.len()-1,c,Direction::Up)]))
            .map(|start| self.traverse_grid(&grid, start))
            .max()
            .unwrap();

        Ok(Box::new(ans))
    }

    fn get_day(&self) -> u8 {
        16
    }
}

impl Day16 {
    /// Returns the next coordinate after taking a step in the given direction
    fn step(&self, r: usize, c: usize, d: Direction) -> (usize, usize, Direction) {
        let (dr, dc) = [(-1,0),(0,1),(1,0),(0,-1)][d as usize];
        ((r as isize + dr) as _, (c as isize + dc) as _, d)
    }
    
    /// Function which traverses through the grid and returns the number of tiles that are energised
    /// in the process
    fn traverse_grid(&self, grid: &[&[u8]], start: (usize,usize, Direction)) -> usize {
        let mut energised = vec![vec![[false; 4]; grid[0].len()]; grid.len()];
        let mut beams = vec![start];

        // Direction change given the current direction as an index
        let right_lean = [Direction::Right, Direction::Up, Direction::Left, Direction::Down];
        let left_lean = [Direction::Left, Direction::Down, Direction::Right, Direction::Up];

        // Trace all of the beams present in the list
        while !beams.is_empty() {
            let mut new_beams = Vec::with_capacity(beams.capacity());
            // Row, col and direction
            for (r,c,d) in beams {
                if r >= grid.len() || c >= grid[0].len() {
                    continue;
                }
                if energised[r][c][d as usize] {
                    continue;
                }
                energised[r][c][d as usize] = true;

                // Trace the new path in a given direction and draw all of the new beams that come from it.
                match (grid[r][c], d) {
                    (b'/',   _) => new_beams.push(self.step(r,c,right_lean[d as usize])),
                    (b'\\',  _) => new_beams.push(self.step(r,c,left_lean[d as usize])),
                    (b'|', Direction::Left|Direction::Right) => new_beams.extend([self.step(r,c,Direction::Up), self.step(r,c,Direction::Down)]),
                    (b'-', Direction::Up|Direction::Down) => new_beams.extend([self.step(r,c,Direction::Left), self.step(r,c, Direction::Right)]),
                    _           => new_beams.push(self.step(r,c,d)),
                }
            }
            beams = new_beams;
        }

        // Get the number of 'energised' tiles now that we have finished processing every beam
        energised.iter().flat_map(|row| row).filter(|x| x.iter().any(|&b| b)).count()
      }
}

/// Test from puzzle input
#[cfg(test)]
mod test {
    use super::*;
    use crate::*;

    #[test]
    fn part1() {
        let challenge = day16::Day16 {};

        //Complete the Challenge
        let answer = challenge
            .part1(
                utils::get_input(challenge.get_day(), utils::InputType::Test1)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "46");
    }

    #[test]
    fn part2() {
        let challenge = day16::Day16 {};

        //Complete the Challenge
        let answer = challenge
            .part2(
                utils::get_input(challenge.get_day(), utils::InputType::Test2)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "51");
    }
}